"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const vscode_1 = require("vscode");
var RoiMode;
(function (RoiMode) {
    RoiMode[RoiMode["Nothing"] = 0] = "Nothing";
    RoiMode[RoiMode["VisibleFiles"] = 1] = "VisibleFiles";
    RoiMode[RoiMode["VisibleLines"] = 2] = "VisibleLines";
    RoiMode[RoiMode["VisibleLinesAndAbove"] = 3] = "VisibleLinesAndAbove";
    RoiMode[RoiMode["OpenFiles"] = 4] = "OpenFiles";
    RoiMode[RoiMode["ProjectFiles"] = 5] = "ProjectFiles";
})(RoiMode = exports.RoiMode || (exports.RoiMode = {}));
class RoiManager {
    constructor(server, documentFilter) {
        this.server = server;
        this.documentFilter = documentFilter;
        this.modeChangedEmitter = new vscode_1.EventEmitter();
        this.onModeChanged = this.modeChangedEmitter.event;
        this.subscriptions = [];
        this.subscriptions.push(vscode_1.window.onDidChangeActiveTextEditor(() => this.send()), vscode_1.window.onDidChangeTextEditorSelection(() => this.send()), vscode_1.window.onDidChangeVisibleTextEditors(() => this.send()), vscode_1.window.onDidChangeTextEditorVisibleRanges(() => this.send()), vscode_1.workspace.onDidOpenTextDocument(() => this.send()), vscode_1.workspace.onDidCloseTextDocument(() => this.send()), server.restarted.on(() => this.send()));
        switch (vscode_1.workspace.getConfiguration('lean').get('roiModeDefault')) {
            case 'nothing':
                this.mode = RoiMode.Nothing;
                break;
            case 'visible':
                this.mode = RoiMode.VisibleFiles;
                break;
            case 'lines':
                this.mode = RoiMode.VisibleLines;
                break;
            case 'linesAndAbove':
                this.mode = RoiMode.VisibleLinesAndAbove;
                break;
            case 'open':
                this.mode = RoiMode.OpenFiles;
                break;
            case 'project':
                this.mode = RoiMode.ProjectFiles;
                break;
            default: this.mode = RoiMode.VisibleLinesAndAbove;
        }
        this.send();
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.roiMode.select', () => __awaiter(this, void 0, void 0, function* () {
            const items = [
                {
                    label: 'nothing',
                    description: 'disable checking',
                    mode: RoiMode.Nothing,
                },
                {
                    label: 'visible lines',
                    description: 'check lines that are currently visible',
                    mode: RoiMode.VisibleLines,
                },
                {
                    label: 'visible lines and above',
                    description: 'check visible lines and the file above it',
                    mode: RoiMode.VisibleLinesAndAbove,
                },
                {
                    label: 'visible files',
                    description: 'check files that are currently visible',
                    mode: RoiMode.VisibleFiles,
                },
                {
                    label: 'open files',
                    description: 'check files that are opened',
                    mode: RoiMode.OpenFiles,
                },
                {
                    label: 'project files',
                    description: 'check files that are in the current workspace',
                    mode: RoiMode.ProjectFiles,
                },
            ];
            const selected = yield vscode_1.window.showQuickPick(items);
            if (selected) {
                this.check(selected.mode);
            }
        })));
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.roiMode.nothing', () => this.check(RoiMode.Nothing)));
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.roiMode.visibleFiles', () => this.check(RoiMode.VisibleFiles)));
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.roiMode.openFiles', () => this.check(RoiMode.OpenFiles)));
        this.subscriptions.push(vscode_1.commands.registerCommand('lean.roiMode.projectFiles', () => this.check(RoiMode.ProjectFiles)));
    }
    compute() {
        return __awaiter(this, void 0, void 0, function* () {
            let paths;
            if (this.mode === RoiMode.ProjectFiles) {
                const files = yield vscode_1.workspace.findFiles('**/*.lean');
                paths = files.map((f) => f.fsPath);
            }
            else {
                paths = vscode_1.workspace.textDocuments.map((d) => d.fileName);
            }
            const visibleRanges = {};
            for (const editor of vscode_1.window.visibleTextEditors) {
                if (vscode_1.languages.match(this.documentFilter, editor.document)) {
                    visibleRanges[editor.document.fileName] =
                        editor.visibleRanges.map((r) => ({
                            begin_line: r.start.line + 1,
                            end_line: r.end.line + 1,
                        }));
                }
            }
            return paths.map((path) => ({ file_name: path, ranges: visibleRanges[path] || [] }));
        });
    }
    modeString() {
        switch (this.mode) {
            case RoiMode.Nothing: return 'nothing';
            case RoiMode.VisibleFiles: return 'visible-files';
            case RoiMode.VisibleLines: return 'visible-lines';
            case RoiMode.VisibleLinesAndAbove: return 'visible-lines-and-above';
            case RoiMode.OpenFiles: return 'open-files';
            case RoiMode.ProjectFiles: return 'open-files';
            default: throw new Error('unknown roi mode');
        }
    }
    send() {
        return __awaiter(this, void 0, void 0, function* () {
            const roi = yield this.compute();
            yield this.server.roi(this.modeString(), roi);
        });
    }
    check(mode) {
        this.mode = mode;
        this.modeChangedEmitter.fire(mode);
        this.send();
    }
    dispose() {
        for (const s of this.subscriptions) {
            s.dispose();
        }
    }
}
exports.RoiManager = RoiManager;
//# sourceMappingURL=roi.js.map